import os
import time

from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import sessionmaker

from .models import Base

DB_TYPE = os.environ.get("DB_TYPE", "sqlite")


def get_pg_conn() -> Engine:
    # Helper function to get environment variable with fallback
    def get_env_var(key: str) -> str:
        task_key = f"KUBICLE_{key}"
        value = os.environ.get(task_key)
        if value is None:
            value = os.environ.get(key)
            if value is None:
                raise ValueError(f"${key} must be set")
        return value

    # Retrieve environment variables with fallbacks
    db_user = get_env_var("DB_USER")
    db_password = get_env_var("DB_PASS")
    db_host = get_env_var("DB_HOST")
    db_name = get_env_var("DB_NAME")

    print(f"\nconnecting to db on postgres host '{db_host}' with db '{db_name}'")
    engine = create_engine(
        f"postgresql+psycopg2://{db_user}:{db_password}@{db_host}/{db_name}",
        client_encoding="utf8",
    )

    return engine


def get_sqlite_conn() -> Engine:
    db_name = os.environ.get("KUBICLE_DB_NAME", "job.db")
    db_path = os.environ.get("KUBICLE_DB_PATH", "./.data")
    db_test = os.environ.get("KUBICLE_DB_TEST", "false") == "true"
    if db_test:
        db_name = f"kubicle_test_{int(time.time())}.db"
    print(f"connecting to local sqlite db ./.data/{db_name}")
    os.makedirs(os.path.dirname(f"{db_path}/{db_name}"), exist_ok=True)
    engine = create_engine(f"sqlite:///{db_path}/{db_name}")
    return engine


if DB_TYPE == "postgres":
    engine = get_pg_conn()
else:
    engine = get_sqlite_conn()
SessionLocal = sessionmaker(bind=engine)

Base.metadata.create_all(bind=engine)


class WithDB:
    @staticmethod
    def get_db():
        """Get a database connection

        Example:
            ```
            for session in self.get_db():
                session.add(foo)
            ```
        """
        db = SessionLocal()
        try:
            yield db
        finally:
            db.close()


def get_db():
    """Get a database connection

    Example:
        ```
        for session in get_db():
            session.add(foo)
        ```
    """
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



import os
from typing import Optional, Dict, List

from agentcore.models import V1UserProfile

from kubicle.runner.base import JobRunner, DEFAULT_OWNER_REF


def DynamicRunner(
    local: bool = False,
    namespace: Optional[str] = None,
    owner_ref: V1UserProfile = DEFAULT_OWNER_REF,
    img: Optional[str] = None,
    envs: Dict[str, str] = {},
    include_env: List[str] = [],
    copy_env: bool = False,
    copy_exclude: Optional[List[str]] = None,
    service_account: str = "default",
) -> JobRunner:
    """
    Returns a JobRunner that dynamically loads a module and function at runtime.
    """

    from kubicle.runner.k8s.runner import K8sJobRunner
    from kubicle.runner.process.runner import ProcessJobRunner

    is_local = os.getenv("JOB_LOCAL")
    if is_local:
        if is_local.lower() == "true":
            local = True

    if local:
        print("using local runner", flush=True)
        return ProcessJobRunner(owner_ref=owner_ref)
    else:
        print("using k8s runner", flush=True)
        return K8sJobRunner(
            namespace=namespace,
            owner_ref=owner_ref,
            img=img,
            envs=envs,
            include_env=include_env,
            copy_env=copy_env,
            copy_exclude=copy_exclude,
            service_account=service_account,
        )



import time
import uuid

from sqlalchemy import Boolean, Column, Float, ForeignKey, Integer, String, Text
from sqlalchemy.orm import declarative_base, relationship

Base = declarative_base()


class JobRecord(Base):
    __tablename__ = "jobs"

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    owner_id = Column(String, nullable=False)
    name = Column(String, nullable=False)
    status = Column(String, nullable=False)
    runtime = Column(String, nullable=False)
    namespace = Column(String, nullable=True)
    phase = Column(String, nullable=True)
    logs = Column(String, nullable=True)
    pid = Column(Integer, nullable=True)
    result = Column(String, nullable=True)
    created = Column(Float, default=time.time)
    updated = Column(Float, default=time.time)
    finished = Column(Float, default=0.0)
    metadata_ = Column(String, nullable=True)



import logging
import sys
import threading
from collections import deque
from time import sleep
from typing import IO

from .base import Job


class StreamCapture:
    def __init__(
        self, stream: IO[str], logger: logging.Logger, log_level: int = logging.INFO
    ) -> None:
        self.original_stream = stream
        self.logger = logger
        self.log_level = log_level

    def write(self, data: str) -> None:
        self.logger.log(self.log_level, data)
        self.original_stream.write(data)

    def flush(self) -> None:
        self.original_stream.flush()


class StreamToLogger:
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(
        self, logger: logging.Logger, log_level: int = logging.INFO, echo: bool = False
    ) -> None:
        self.logger: logging.Logger = logger
        self.log_level: int = log_level
        self.linebuf: str = ""
        self.echo: bool = echo

    def write(self, buf: str) -> None:
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())
            # if self.echo:
            #     print(line.rstrip())

        self.flush()

    def flush(self) -> None:
        pass


class BufferedJobLogHandler(logging.Handler):
    def __init__(self, job: Job, flush_interval=5, buffer_size=10):
        super().__init__()
        self.job = job
        self.flush_interval = flush_interval
        self.buffer_size = buffer_size
        self.buffer = deque()
        self.lock = threading.Lock()
        self.flusher_thread = threading.Thread(
            target=self.flush_periodically, daemon=True
        )
        self.flusher_thread.start()

    def emit(self, record):
        with self.lock:  # type: ignore
            self.buffer.append(self.format(record))
            if len(self.buffer) >= self.buffer_size:
                self.flush_buffer()

    def flush_buffer(self):
        with self.lock:  # type: ignore
            while self.buffer:
                log_entry = self.buffer.popleft()
                # Instead of printing, append to the Job's logs
                self.job.log(log_entry)

    def flush_periodically(self):
        while True:
            sleep(self.flush_interval)
            if self.buffer:
                self.flush_buffer()



from typing import Dict, List, Optional

from pydantic import BaseModel


class V1Job(BaseModel):
    id: str
    owner_id: str
    status: str
    runtime: str
    name: str
    namespace: Optional[str] = None
    logs: Optional[str] = None
    result: Optional[str] = None
    pid: Optional[int] = None
    created: float
    updated: float
    finished: float
    metadata: Dict[str, str] = {}


class V1Jobs(BaseModel):
    jobs: List[V1Job]



from abc import ABC, abstractmethod
from typing import Callable, TypeVar

from pydantic import BaseModel
from agentcore.models import V1UserProfile

from kubicle.base import Job

InputType = TypeVar("InputType", bound=BaseModel)
OutputType = TypeVar("OutputType", bound=BaseModel)


class JobRunner(ABC):
    """
    A base class for job runners.
    """

    @abstractmethod
    def run(self, fn: Callable[[InputType], OutputType], input: InputType) -> Job:
        pass

    @abstractmethod
    def refresh(self, job_id: str) -> Job:
        pass

    @abstractmethod
    def cancel(self, job_id: str) -> Job:
        pass


DEFAULT_OWNER_REF = V1UserProfile(email="anonymous@agentsea.ai")



import json
import time
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Dict, List, Optional

import shortuuid

from kubicle.db.conn import WithDB
from kubicle.db.models import JobRecord
from kubicle.server.models import V1Job


class JobStatus(Enum):
    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"
    CANCELED = "canceled"
    UNKNOWN = "unknown"


class JobRuntime(Enum):
    K8s = "k8s"
    Process = "process"


@dataclass
class Job(WithDB):
    """A backgound job"""

    owner_id: str
    status: JobStatus
    runtime: str
    name: str
    id: str = field(default_factory=lambda: str(shortuuid.uuid()))
    namespace: Optional[str] = None
    logs: Optional[str] = None
    result: Optional[str] = None
    pid: Optional[int] = None
    created: float = field(default_factory=lambda: time.time())
    updated: float = field(default_factory=lambda: time.time())
    finished: float = field(default_factory=lambda: 0.0)
    metadata: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.save()

    def to_record(self) -> JobRecord:
        metadata_serialized = json.dumps(self.metadata)
        return JobRecord(
            id=self.id,
            owner_id=self.owner_id,
            status=self.status.value,
            name=self.name,
            runtime=self.runtime,
            namespace=self.namespace,
            logs=self.logs,
            pid=self.pid,
            result=self.result,
            created=self.created,
            updated=self.updated,
            finished=self.finished,
            metadata_=metadata_serialized,
        )

    @classmethod
    def from_record(cls, record: JobRecord) -> "Job":
        obj = cls.__new__(cls)
        obj.id = str(record.id)
        obj.owner_id = str(record.owner_id)
        obj.name = str(record.name)
        obj.status = JobStatus(str(record.status))
        obj.runtime = str(record.runtime)
        obj.namespace = record.namespace  # type: ignore
        obj.logs = record.logs  # type: ignore
        obj.pid = record.pid  # type: ignore
        obj.result = str(record.result)
        obj.created = record.created  # type: ignore
        obj.updated = record.updated  # type: ignore
        obj.finished = record.finished  # type: ignore
        obj.metadata = json.loads(str(record.metadata_)) if record.metadata_ else {}  # type: ignore
        return obj

    def to_v1(self) -> "V1Job":
        return V1Job(**asdict(self))

    @classmethod
    def from_v1(cls, schema: "V1Job") -> "Job":
        obj = cls.__new__(cls)
        obj.id = schema.id
        obj.owner_id = schema.owner_id
        obj.name = schema.name
        obj.status = JobStatus(schema.status)
        obj.runtime = schema.runtime
        obj.namespace = schema.namespace
        obj.logs = schema.logs
        obj.pid = schema.pid
        obj.result = schema.result
        obj.created = schema.created
        obj.updated = schema.updated
        obj.finished = schema.finished
        obj.metadata = schema.metadata
        return obj

    def log(self, message: str, add_timestamp=True) -> None:
        """Appends a log message to the job's logs, with an optional timestamp."""
        if add_timestamp:
            timestamp = time.strftime("[%Y-%m-%d %H:%M:%S]", time.gmtime())
            log_entry = f"{timestamp} {message}\n"
        else:
            log_entry = message + "\n"  # Assuming message is already formatted

        if not self.logs:
            self.logs = log_entry
        else:
            self.logs += log_entry

        self.updated = time.time()
        self.save()

    def save(self) -> None:
        for db in self.get_db():
            db.merge(self.to_record())
            db.commit()

    @classmethod
    def find(cls, **kwargs) -> List["Job"]:
        for db in cls.get_db():
            records = db.query(JobRecord).filter_by(**kwargs).all()
            return [cls.from_record(record) for record in records]

        raise SystemError("no session")

    @classmethod
    def delete(cls, id: str) -> None:
        for db in cls.get_db():
            record = db.query(JobRecord).filter_by(id=id).first()
            if record:
                db.delete(record)
                db.commit()

    def refresh(self) -> None:
        """
        Refreshes the instance's state based on the latest data from the database.
        """
        for db in self.get_db():
            record = db.query(JobRecord).filter_by(id=self.id).first()
            if not record:
                raise ValueError(f"No job found with id {self.id}")

            # Refreshing the instance attributes from the record
            self.owner_id = str(record.owner_id)
            self.name = str(record.name)
            self.status = JobStatus(str(record.status))
            self.runtime = str(record.runtime)
            self.namespace = record.namespace  # type: ignore
            self.logs = record.logs  # type: ignore
            self.result = str(record.result)
            self.created = record.created  # type: ignore
            self.updated = record.updated  # type: ignore
            self.finished = record.finished  # type: ignore
            self.metadata = (
                json.loads(str(record.metadata_)) if record.metadata_ else {}  # type: ignore
            )



from multiprocessing import Process
import time
import importlib
import inspect
import traceback
from typing import Callable
import os
import signal

from agentcore.models import V1UserProfile

from kubicle.base import Job, JobStatus, JobRuntime
from kubicle.runner.base import JobRunner, InputType, OutputType, DEFAULT_OWNER_REF
from kubicle.util import get_random_characters


def job_wrapper(module_name: str, function_name: str, input: InputType, job_id: str):  # type: ignore
    print("job wrapper..", flush=True)
    try:
        module = importlib.import_module(module_name)
        fn: Callable[[InputType], OutputType] = getattr(module, function_name)  # type: ignore
    except (ModuleNotFoundError, AttributeError) as e:
        error_message = f"Error loading function: {e}\n{traceback.format_exc()}"
        print(error_message, flush=True)
        save_job_error(job_id, error_message)
        return

    jobs = Job.find(id=job_id)
    if not jobs:
        error_message = f"Job {job_id} not found"
        print(error_message, flush=True)
        save_job_error(job_id, error_message)
        return
    job = jobs[0]

    try:
        print("calling function", flush=True)
        output = fn(input)
        print("function returned", flush=True)
        job.status = JobStatus.FINISHED
        job.result = output.model_dump_json()
        job.save()
    except Exception as e:
        error_message = (
            f"Error during function execution: {e}\n{traceback.format_exc()}"
        )
        print(error_message, flush=True)
        save_job_error(job_id, error_message)
    finally:
        job.refresh()
        job.finished = time.time()
        job.save()


def save_job_error(job_id: str, error_message: str):
    print("saving error...", flush=True)
    jobs = Job.find(id=job_id)
    if jobs:
        job = jobs[0]
        job.status = JobStatus.FAILED
        job.result = error_message
        job.finished = time.time()
        job.save()
        print("saved error", flush=True)


class ProcessJobRunner(JobRunner):
    """
    Process job runner
    """

    def __init__(self, owner_ref: V1UserProfile = DEFAULT_OWNER_REF) -> None:
        self.owner = owner_ref
        print("ProcessJobRunner.init()", flush=True)

    def run(self, fn: Callable[[InputType], OutputType], input: InputType) -> Job:
        print("ProcessJobRunner.run()", flush=True)
        run_fn_name = fn.__name__
        run_module = inspect.getmodule(fn).__name__  # type: ignore

        print(f"RUN_MODULE: {run_module}", flush=True)
        print(f"RUN_FN: {run_fn_name}", flush=True)
        print(f"INPUT: {input}", flush=True)
        print(f"OWNER: {self.owner}", flush=True)

        job_name = f"{run_fn_name}-{get_random_characters()}".lower().replace("_", "-")

        if not self.owner.email:
            raise Exception("Owner email is not set")

        job = Job(
            self.owner.email,
            JobStatus.RUNNING,
            JobRuntime.Process.value,
            job_name,
        )

        print("starting process...", flush=True)

        p = Process(
            target=job_wrapper,
            args=(run_module, run_fn_name, input, job.id),
        )
        p.daemon = True
        p.start()

        job.pid = p.pid
        job.save()

        return job

    def refresh(self, job_id: str) -> Job:
        # Find the job in the database or similar
        found = Job.find(id=job_id)
        if not found:
            raise ValueError(f"Job {job_id} not found")
        job = found[0]

        # Check if we have a stored PID and if the process is still running
        if job.pid:
            try:
                os.kill(job.pid, 0)  # Check if process is alive
                job.status = JobStatus.RUNNING
            except OSError:
                job.refresh()  # refresh the job status from the database
                if job.status != JobStatus.FINISHED:
                    job.status = JobStatus.FAILED
        else:
            job.status = JobStatus.FAILED

        job.save()
        return job

    def cancel(self, job_id: str) -> Job:
        # Find the job in the database or similar
        found = Job.find(id=job_id)
        if not found:
            raise ValueError(f"Job {job_id} not found")
        job = found[0]

        # Check if we have a stored PID and if the process is still running
        if job.pid:
            try:
                print(
                    f"Terminating process with PID {job.pid} for job {job_id}",
                    flush=True,
                )
                os.kill(job.pid, signal.SIGTERM)  # Send termination signal
                job.status = JobStatus.CANCELED
            except OSError:
                job.status = JobStatus.FAILED  # Process already finished or not found
        else:
            job.status = JobStatus.FAILED

        job.finished = time.time()
        job.save()

        return job
